module Sheet where

import Data.Array

type CellRef = (Char, Int)

type Sheet a = Array CellRef a

data BinOp = Add | Sub | Mul | Div

data Exp = Lit Double
         | Ref CellRef
         | App BinOp Exp Exp
         | Sum CellRef CellRef
         | Avg CellRef CellRef

evalExp :: Sheet Double -> Exp -> Double
evalExp _ (Lit v)        = v
evalExp s (Ref r)        = s ! r
evalExp s (App op e1 e2) = (evalOp op) (evalExp s e1) (evalExp s e2)
evalExp s (Sum r1 r2)    = sum (cellsInRange s r1 r2)
evalExp s (Avg r1 r2)    = sum xs / fromIntegral (length xs)
    where
        xs = cellsInRange s r1 r2

cellsInRange :: Sheet a -> CellRef -> CellRef -> [a]
cellsInRange s r1 r2 = [ s ! r | r <- range (r1, r2) ]

-- Alternatively, if we wish to be more "forgiving" regarding how ranges
-- are defined:
--
-- cellsInRange :: Sheet a -> CellRef -> CellRef -> [a]
-- cellsInRange s (x1,y1) (x2,y2) = [ s ! r | r <- range (r1, r2) ]
--     where
--         r1 = ((min x1 x2), (min y1 y2))
--         r2 = ((max x1 x2), (max y1 y2))

evalOp :: BinOp -> (Double -> Double -> Double)
evalOp Add = (+)
evalOp Sub = (-)
evalOp Mul = (*)
evalOp Div = (/)


evalSheet :: Sheet Exp -> Sheet Double
evalSheet s = s'
    where
        s' = array (bounds s) [ (r, evalExp s' (s ! r)) | r <- indices s ]

testSheet :: Sheet Exp
testSheet = array (('a', 1), ('e', 3))
                  [ (('a', 1), Lit 1.0),
                    (('a', 2), Ref ('b', 1)),
                    (('a', 3), Lit 3.0),
                    (('b', 1), App Add (Ref ('c', 2)) (Ref ('b', 2))),
                    (('b', 2), Lit 2.0),
                    (('b', 3), App Mul (Ref ('a', 1)) (Ref ('a', 2))),
                    (('c', 1), App Add (Ref ('a',1)) (Ref ('b',3))),
                    (('c', 2), Lit 3.0),
                    (('c', 3), Lit 7.0),
                    (('d', 1), Sum ('a', 1) ('c', 1)),
                    (('d', 2), Sum ('a', 2) ('c', 2)),
                    (('d', 3), Sum ('a', 3) ('c', 3)),
                    (('e', 1), Avg ('a', 1) ('c', 1)),
                    (('e', 2), Avg ('a', 2) ('c', 2)),
                    (('e', 3), Avg ('a', 3) ('c', 3))
                  ]
